////////////////////////////////////////////////////////////////////////////////
//
// KS-10 Processor
//
// Brief
//   UBA Loopback
//
// Details
//   This module implements the UBA looback that is used to test the UBA
//   adapter.
//
// File
//   ubaloop.v
//
// Author
//   Rob Doyle - doyle (at) cox (dot) net
//
////////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2012-2014 Rob Doyle
//
// This source file may be used and distributed without restriction provided
// that this copyright statement is not removed from the file and that any
// derivative work contains the original copyright notice and the associated
// disclaimer.
//
// This source file is free software; you can redistribute it and/or modify it
// under the terms of the GNU Lesser General Public License as published by the
// Free Software Foundation; version 2.1 of the License.
//
// This source is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
// for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this source; if not, download it from
// http://www.gnu.org/licenses/lgpl.txt
//
////////////////////////////////////////////////////////////////////////////////

`default_nettype none
`include "ubapage.vh"
`include "../cpu/bus.vh"

  module UBALOOP(clk, rst, busADDRI, busADDRO, busDATAI, busACKI, pageFLAGS,
                 loopREAD, loopWRITE, loopREQO, loopADDRO, loopDATAO);

   input         clk;                           // Clock
   input         rst;                           // Reset
   input  [0:35] busADDRI;                      // Backplane bus address in
   input  [0:35] busADDRO;                      // Backplane bus address out
   input  [0:35] busDATAI;                      // Backplane bus data in
   input         busACKI;                       // Backplane bus acknowledge
   input  [0: 3] pageFLAGS;                     // Pager flags
   input         loopREAD;                      // Read from loop around
   input         loopWRITE;                     // Write to loop around
   output        loopREQO;                      // Loopback Request
   output [0:35] loopADDRO;                     // Loopback Address
   output [0:35] loopDATAO;                     // Loopback Address

   //
   // Read and Write Flags
   //

   localparam [0:17] readFLAGS  = 18'o040000;
   localparam [0:17] writeFLAGS = 18'o010000;

   //
   // States
   //

   localparam [0:2] stateIDLE      = 0,
                    stateADDRWAIT  = 1,
                    stateREADWAIT  = 2,
                    stateREADREQ   = 3,
                    statePAUSE     = 4,
                    stateWRITEWAIT = 5,
                    stateWRITEREQ  = 6;

   //
   // Address Bus
   //

   wire busIO = `busIO(busADDRI);

   //
   // Page Flags
   //

   wire flagsRPW = `flagsRPW(pageFLAGS);
   wire flagsFTM = `flagsFTM(pageFLAGS);

   //
   // Capture Address and Data
   //

   reg [18:35] loopADDR;
   reg [ 0:35] loopDATA;

   always @(posedge clk or posedge rst)
     begin
        if (rst)
          begin
             loopADDR <= 0;
             loopDATA <= 0;
          end
        else
          begin
             if (loopWRITE)
               begin
                  loopADDR <= busADDRI[18:35];
                  if (busADDRI[34] == 0)
                    loopDATA[ 0:17] <= busDATAI[18:35];         // Even word
                  else
                    loopDATA[18:35] <= busDATAI[18:35];         // Odd word
               end
             if (loopREAD)
               loopADDR <= busADDRI[18:35];
          end
     end

   //
   // Write Loopback State Machine
   //

   reg [0: 2] wrstate;
   reg [0:35] loopDATAO;

   always @(posedge clk or posedge rst)
     begin
        if (rst)
          begin
             loopDATAO <= 0;
             wrstate   <= stateIDLE;
          end
        else
          case (wrstate)

            //
            // Wait a clock cycle.  It takes a clock cycle for the address to
            // propigate through the paging RAM.
            //

            stateIDLE:
              if (loopWRITE)
                wrstate <= stateADDRWAIT;

            //
            // Figure out what kind of DMA cycle to generate.
            //  Some types generate a Read/Pause/Write cycle
            //  Some types just generate a write cycle
            //

            stateADDRWAIT:
                begin

                   //
                   // FTM even word is just buffered locally, not written.
                   //

                   if (flagsFTM & (loopADDR[34] == 0))
                     wrstate <= stateIDLE;

                   //
                   // FTM odd word is combined with buffered even word.
                   //

                   else if (flagsFTM & (loopADDR[34] == 1))
                     begin
                        loopDATAO <= loopDATA;
                        wrstate   <= stateWRITEWAIT;
                     end

                   //
                   // Even word (non RPW)
                   //

                   else if (!flagsRPW & (loopADDR[34] == 0))
                     begin
                        loopDATAO <= {loopDATA[0:17], busADDRO[18:35]};
                        wrstate   <= stateWRITEWAIT;
                     end

                   //
                   // Everything else is a RPW cycle
                   //

                   else
                     wrstate <= stateREADWAIT;
                end

            //
            // Wait to generate a bus requests for a read cycle
            //

            stateREADWAIT:
              if (!busIO)
                wrstate <= stateREADREQ;

            //
            // Wait for a bus acknowledge
            //

            stateREADREQ:
              if (busACKI)
                begin

                   //
                   // Even word RPW
                   //

                   if (loopADDR[34] == 0)
                     loopDATAO <= {loopDATA[ 0:17], busDATAI[18:35]};

                   //
                   // Odd word RPW
                   //

                   if (loopADDR[34] == 1)
                     loopDATAO <= {busDATAI[ 0:17], loopDATA[18:35]};

                   wrstate <= statePAUSE;
                end

            //
            // Pause
            //

            statePAUSE:
              wrstate <= stateWRITEWAIT;

            //
            // Wait to generate a bus requests for a write cycle
            //

            stateWRITEWAIT:
              if (!busIO)
                wrstate <= stateWRITEREQ;

            //
            // Wait for a bus acknowledge
            //

            stateWRITEREQ:
              if (busACKI)
                wrstate <= stateIDLE;

          endcase
     end

   //
   // Read Loopback State Machine
   //

   reg [0: 2] rdstate;
   reg [0:35] loopDATAI;

   always @(posedge clk or posedge rst)
     begin
        if (rst)
          begin
             loopDATAI <= 0;
             rdstate   <= stateIDLE;
          end
        else

          case (rdstate)

            stateIDLE:
              if (loopREAD)
                rdstate <= stateADDRWAIT;

            stateADDRWAIT:
                rdstate <= stateREADWAIT;

            stateREADWAIT:
              if (!busIO)
                rdstate <= stateREADREQ;

            stateREADREQ:
              if (busACKI)
                begin
                   rdstate <= stateIDLE;
                end

          endcase
     end


   //
   // Generate Address flags for the various bus cycles
   //

   reg [0:35] loopADDRO;
   always @*
     begin
        if ((wrstate == stateREADWAIT) |
            (wrstate == stateREADREQ)  |
            (rdstate == stateREADWAIT) |
            (rdstate == stateREADREQ))
          loopADDRO <= {readFLAGS, loopADDR};
        else
          loopADDRO <= {writeFLAGS, loopADDR};
     end

   //
   // Outputs
   //

   assign loopREQO = ((wrstate == stateREADREQ) |
                      (wrstate == stateWRITEREQ)|
                      (rdstate == stateREADREQ) |
                      (rdstate == stateWRITEREQ));

endmodule
